import math
import random

import pandas

import actuarial_tables as at
import financial_math as fm


def t_p_x(x, t, table):
    lower_t = math.floor(t); dt = t - lower_t
    upper_x = math.ceil(x); dx = upper_x - x
    px = table['px']; qx = table['qx']; prob = 1
    if (x+t) > len(px):
        return 0
    elif t == 0:
        return 1
    else:
        if (dt==0) and (dx==0):
            x=int(x); t=int(t);
            for i in px[x:x+t]:
                prob = prob*i
            return prob
        elif (dt==0) and (dx!=0):
            t=int(t)
            return t_p_x(upper_x, t, table)*(1-dx*table['qx'][math.floor(x)])
        elif (dt!=0) and (dx==0):
            x=int(x)
            if x+t < x+1:
                return 1-t*table['qx'][x]
            if x+t > x+1:
                return t_p_x(x, math.floor(x+t)-x, table)*t_p_x(math.floor(x+t), x+t-math.floor(x+t), table)
        else:
            if x+t < upper_x:
                return 1-t*table['qx'][math.floor(x)]
            elif x+t > upper_x:
                return (1-dx*table['qx'][math.floor(x)])*(t_p_x(upper_x, x+t-upper_x, table))
            else:
                return 1-t*table['qx'][math.floor(x)]
            
def t_q_x(x, t, table):
    return 1-t_p_x(x,t,table)

def find_premium(x, t, benefits, interest_rates, table, premium_time_points, loss_bound = 0.5, loss_target = 0, start_premium = 0, dp = 0.25, max_iter = 10000):
    insurance = TermLifeInsurance(x, t, benefits, interest_rates, table)
    if start_premium == 0:
        premium = dp
    else:
        premium = start_premium
    annuity = TermLifeAnnuity(x, t, [Premium(i, premium) for i in premium_time_points], interest_rates, table)
    loss = insurance.calculate_apv() - annuity.calculate_apv()
    iter_num = 0
    while abs(loss-loss_target) > abs(loss_bound):
        if iter_num == max_iter:
            return None
        premium = premium + dp
        annuity.payments = [Premium(i, premium) for i in premium_time_points]
        annuity.adjustments()
        loss = insurance.calculate_apv()-annuity.calculate_apv()
        iter_num += 1
    print("Expected Loss = {}".format(loss))
    return premium

def combined_two_cashflows(cashflow_df1, cashflow_df2, interest_rates, collective_interest_rates=None):
    if collective_interest_rates == None:
        collective_interest_rates = [fm.InterestRate(0, 0, 'annual')]
    cols = ['profit_exclude_collective_fund', 'income', 'expense', 'income_from_collective_fund']
    t_pols = sorted(list(set(list(cashflow_df1.t_policy) + list(cashflow_df2.t_policy))))
    df = pandas.DataFrame(t_pols, columns=['t_policy'])
    for col in cols:
        results = []
        for t in t_pols:
            result1 = cashflow_df1[cashflow_df1.t_policy==t][col]
            result2 = cashflow_df2[cashflow_df2.t_policy==t][col]
            if len(result1)==0:
                result1 = 0
            else:
                result1 = result1.values[0]
            if len(result2)==0:
                result2 = 0
            else:
                result2 = result2.values[0]
            results.append(result1+result2)
            prev_results = (result1, result2)
        df[col] = pandas.Series(results)
    exp_contributions = [fm.Contribution(df.t_policy[i], df.profit_exclude_collective_fund[i]) for i in range(len(df))]
    exp_contributions_fund = [fm.Contribution(df.t_policy[i], df.income_from_collective_fund[i]) for i in range(len(df))]
    accumulated = []
    for i in range(1, len(exp_contributions)+1):
        involved = exp_contributions[0:i]
        total = sum([j.accumulate(involved[-1].t, interest_rates)  for j in involved])
        accumulated.append(total)
    accumulated_2 = []
    for i in range(1, len(exp_contributions_fund)+1):
        involved = exp_contributions_fund[0:i]
        total = sum([j.accumulate(involved[-1].t, collective_interest_rates)  for j in involved])
        accumulated_2.append(total)
    df['accumulated_profit_exclude_collective_fund'] = pandas.Series(accumulated, index = df.index)
    df['accumulated_income_collective_fund'] = pandas.Series(accumulated_2, index = df.index)
    df['total_accumulated_profit'] = df['accumulated_profit_exclude_collective_fund'] + df['accumulated_income_collective_fund']
    return df

def combined_cashflows(cashflows, interest_rates, collective_interest_rates = None):
    df1 = cashflows[0]; df2 = cashflows[1]
    df = combined_two_cashflows(df1, df2, interest_rates, collective_interest_rates)
    for cf in cashflows[2:]:
        df = combined_two_cashflows(df, cf, interest_rates, collective_interest_rates)
    return df


class Benefit(fm.Contribution):

    def __init__(self, t, amount, endowment = False):
        super().__init__(t, amount)
        self.t = math.ceil(t)
        self.endowment = endowment

    def claim_prob(self, x, table):
        if self.endowment:
            return t_p_x(x, max(self.t-x, 0), table) 
        return t_p_x(x, max(self.t-x-1, 0), table)*t_q_x(x + max(self.t-x-1, 0), self.t - x - max(self.t-x-1, 0), table)

    def expected_at(self, x, table, interest_rates):
        if self.t > x:
            return self.claim_prob(x, table)*self.value_at(x, interest_rates)
        else:
            return 0


class Premium(fm.Contribution):

    def __init__(self, t, amount, collective_fund_proportion = 0):
        super().__init__(t, amount)
        self.t = t
        self.original_amount = amount
        self._collective_fund_premium = None
        self.original()
        if (0<collective_fund_proportion<=1):
            self.collective_fund_proportion = collective_fund_proportion
            self.partial_transform_to_collective_fund(self.collective_fund_proportion)

    def multiplied(self, factor):
        if factor >0:
            return self.amount*factor

    def added(self, amount):
        return self.amount + amount
            
    def payment_prob(self, x, table, next_payment_time = None):
        dt = math.ceil(self.t) - self.t
        if dt == 0:
            self.t = int(self.t)
        if next_payment_time == None:
            return t_p_x(x, self.t-x, table)
        else:
            return t_p_x(x, self.t-x, table)*t_q_x(self.t, next_payment_time-self.t, table)
        
    def expected_at(self, x, table, interest_rates):
        return self.payment_prob(x, table)*(self.value_at(x, interest_rates))

    def collective_fund_premium(self, updated=True):
        if updated:
            self._collective_fund_premium = Premium(self.t, self.transformed_amount)
        else:
            if self._collective_fund_premium == None:
                self._collective_fund_premium = Premium(self.t, 0)
            else:
                pass
        return self._collective_fund_premium

    def partial_transform_to_collective_fund(self, percent):
        if percent > 1 or percent < 0:
            return None
        self.original()
        self.collective_fund_proportion = percent
        self.transformed_amount = self.collective_fund_proportion*self.amount
        self.amount -= self.transformed_amount
    
    def original(self):
        self.amount = self.original_amount
        self.transformed_amount = 0
        self.collective_fund_proportion = 0
        

class TermLifeInsurance(object):

    def __init__(self, x, term, benefits, interest_rates, table):
        self.x = math.floor(x)
        self.t_start = self.x
        self.term = math.floor(term)
        self.t_end = self.x + self.term
        self.table = table
        self.benefits = benefits
        self.interest_rates = interest_rates
        self.adjustments()
        
    def adjustments(self):
        benefits = []; 
        for i in range(len(self.benefits)):
            if (self.x < self.benefits[i].t <= self.t_end):
                benefits.append( self.benefits[i] )
        self.benefits = benefits
        self.benefits = sorted(self.benefits, key = lambda x: x.t)

        int_rates = []; t_taken = []
        for i in range(len(self.interest_rates)):
            if self.interest_rates[i].t not in t_taken:
                int_rates.append( self.interest_rates[i] )
                t_taken.append(self.interest_rates[i].t)
        self.interest_rates = int_rates
        if 0 not in t_taken:
            self.interest_rates[0].t = 0
        self.interest_rates = sorted(self.interest_rates, key = lambda x: x.t)

    def expected_expense(self, t_end):
        if t_end < self.x:
            return 0
        else:
            benefits = [b for b in self.benefits \
                        if self.x <= b.t <= t_end]
            return {'t': [b.t for b in benefits], \
                    'amount': [b.amount for b in benefits], \
                    'probability': [b.claim_prob(self.x, self.table) for b in benefits], \
                    'object': benefits, \
                    'expected': [b.amount*b.claim_prob(self.x, self.table) for b in benefits] \
                    }

    def expected_expense_df(self, t_end):
        expense = self.expected_expense(t_end)
        expense_df = pandas.DataFrame([(i,j) for i,j in zip(expense['t'], expense['expected'])], columns = ["t", "expense"])
        return expense_df

    def acc_expected_expense(self, t_end, interest_rates):
        if t_end < self.x:
            return 0
        else:
            expected_expense = self.expected_expense(t_end)
            n = len(expected_expense['t'])
            return sum([expected_expense['probability'][i]*expected_expense['object'][i].accumulate(t_end, interest_rates) \
                        for i in range(n)])
        
    def calculate_apv(self, point = "issue"):
        if point == "issue" or point < self.x:
            point = self.x
        return sum([b.expected_at(point, self.table, self.interest_rates) for b in self.benefits])

    def minimum_level_premium(self, annuity = 'immediate', interval = "annual", point = "issue"):
        if point == "issue":
            point = self.x
        exp = []
        if interval == "annual":
            dt = 1
        elif interval == "semi-annual":
            dt = 1/2
        elif interval == "quarter":
            dt = 1/4
        elif interval == "month":
            dt = 1/12
        n_payments = round((self.t_end - point)/dt)
        if annuity == 'immediate':
            for t in range(1, n_payments+1):
                payments =  [Premium(point+dt*(i+1), 1) for i in range(t)]
                payments_pv = sum([p.value_at(point, self.interest_rates) for p in payments])
                if t==n_payments:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table))
                else:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table, next_payment_time = point+dt*(t+1)))
        elif annuity == 'due':
            for t in range(1, n_payments+1):
                payments =  [Premium(point + dt*i, 1) for i in range(t)]
                payments_pv = sum([p.value_at(point, self.interest_rates) for p in payments])
                if t==n_payments:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table))
                else:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table, next_payment_time = point+dt*t))
        EXP = sum(exp)
        return self.calculate_apv(point)/EXP


class PureEndowmentInsurance(object):

    def __init__(self, x, term, benefit, interest_rates, table):
        self.x = math.floor(x)
        self.t_start = self.x
        self.term = math.floor(term)
        self.t_end = self.x + self.term
        self.table = table
        benefit.endowment = True
        self.benefit = [benefit]
        self.interest_rates = interest_rates
        self.adjustments()
        
    def adjustments(self):
        int_rates = []; t_taken = []
        for i in range(len(self.interest_rates)):
            if self.interest_rates[i].t not in t_taken:
                int_rates.append( self.interest_rates[i] )
                t_taken.append(self.interest_rates[i].t)
        self.interest_rates = int_rates
        if 0 not in t_taken:
            self.interest_rates[0].t = 0
        self.interest_rates = sorted(self.interest_rates, key = lambda x: x.t)

    def expected_expense(self, t_end):
        if t_end < self.x:
            return 0
        else:
            benefit = [b for b in self.benefit \
                        if self.x <= b.t <= t_end]
            return {'t': [b.t for b in benefit], \
                    'amount': [b.amount for b in benefit], \
                    'probability': [b.claim_prob(self.x, self.table) for b in benefit], \
                    'object': benefit, \
                    'expected': [b.amount*b.claim_prob(self.x, self.table) for b in benefit] \
                    }

    def expected_expense_df(self, t_end):
        expense = self.expected_expense(t_end)
        expense_df = pandas.DataFrame([(i,j) for i,j in zip(expense['t'], expense['expected'])], columns = ["t", "expense"])
        return expense_df

    def acc_expected_expense(self, t_end, interest_rates):
        if t_end < self.x:
            return 0
        else:
            expected_expense = self.expected_expense(t_end)
            n = len(expected_expense['t'])
            return sum([expected_expense['probability'][i]*expected_expense['object'][i].accumulate(t_end, interest_rates) \
                        for i in range(n)])
        
    def calculate_apv(self, point = "issue"):
        if point == "issue" or point < self.x:
            point = self.x
        return self.benefit[0].expected_at(point, self.table, self.interest_rates)

    def minimum_level_premium(self, annuity = 'immediate', interval = "annual", point = "issue"):
        if point == "issue":
            point = self.x
        exp = []
        if interval == "annual":
            dt = 1
        elif interval == "semi-annual":
            dt = 1/2
        elif interval == "quarter":
            dt = 1/4
        elif interval == "month":
            dt = 1/12
        n_payments = round((self.t_end - point)/dt)
        if annuity == 'immediate':
            for t in range(1, n_payments+1):
                payments =  [Premium(point+dt*(i+1), 1) for i in range(t)]
                payments_pv = sum([p.value_at(point, self.interest_rates) for p in payments])
                if t==n_payments:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table))
                else:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table, next_payment_time = point+dt*(t+1)))
        elif annuity == 'due':
            for t in range(1, n_payments+1):
                payments =  [Premium(point + dt*i, 1) for i in range(t)]
                payments_pv = sum([p.value_at(point, self.interest_rates) for p in payments])
                if t==n_payments:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table))
                else:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table, next_payment_time = point+dt*t))
        EXP = sum(exp)
        return self.calculate_apv(point)/EXP


class TermLifeAnnuity(object):

    def __init__(self, x, term, payments, interest_rates, table):
        self.x = math.floor(x)
        self.t_start = self.x
        self.term = math.floor(term)
        self.table = table
        self.payments = payments
        self.interest_rates = interest_rates
        self.adjustments()
        
    def adjustments(self):
        payments = []; 
        for i in range(len(self.payments)):
            if (self.x <= self.payments[i].t <= self.x + self.term):
                payments.append( self.payments[i] )
        self.payments = payments
        self.payments = sorted(self.payments, key = lambda x: x.t)

        int_rates = []; t_taken = []
        for i in range(len(self.interest_rates)):
            if self.interest_rates[i].t not in t_taken:
                int_rates.append( self.interest_rates[i] )
                t_taken.append(self.interest_rates[i].t)
        self.interest_rates = int_rates
        if 0 not in t_taken:
            self.interest_rates[0].t = 0
        self.interest_rates = sorted(self.interest_rates, key = lambda x: x.t)

    def expected_payments(self, t_end):
        if t_end < self.x:
            return 0
        else:
            payments = [p for p in self.payments \
                        if self.x <= p.t <= t_end]
            return {'t': [p.t for p in payments], \
                    'amount': [p.amount for p in payments], \
                    'probability': [p.payment_prob(self.x, self.table) for p in payments], \
                    'object': payments, \
                    'expected': [p.amount*p.payment_prob(self.x, self.table) for p in payments] \
                    }

    def expected_payments_df(self, t_end):
        income = self.expected_payments(t_end)        
        income_df = pandas.DataFrame([(i,j) for i,j in zip(income['t'], income['expected'])], columns = ["t", "income"])
        return income_df

    def acc_expected_payments(self, t_end, interest_rates):
        if t_end < self.x:
            return 0
        else:
            expected_payments = self.expected_payments(t_end)
            n = len(expected_payments['t'])
            return sum([expected_payments['probability'][i]*expected_payments['object'][i].accumulate(t_end, interest_rates) \
                        for i in range(n)])
            
    def calculate_apv(self, point = "issue"):
        if point == "issue" or point < self.x:
            point = self.x
        exp = []
        np = len(self.payments)
        for i in range(np):
            if self.payments[i].t < point:
                continue
            payments = [p for p in self.payments[0:(i+1)] if point <= p.t]
            payments_pv = sum([p.value_at(point, self.interest_rates) for p in payments])
            if len(payments) > 0:
                if i < np - 1:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table, next_payment_time = self.payments[i+1].t))
                else:
                    exp.append(payments_pv*payments[-1].payment_prob(point, self.table))
        return sum(exp)


class ActuarialModel(object):

    def __init__(self, insurance, annuity, policy_year, policy_month=1):
        self.insurance = insurance
        self.interest_rates = self.insurance.interest_rates
        self.policy_year = int(max(0, policy_year))
        self.policy_month = int(max(1, min(policy_month, 12)))
        self.t_policy = self.policy_year + self.policy_month/12
        if annuity[1] == "annual":
            dt = 1
        elif annuity[1] == "semi-annual":
            dt = 1/2
        elif annuity[1] == "quarter":
            dt = 1/4
        elif annuity[1] == "month":
            dt = 1/12
        n_payments = round((self.insurance.term)/dt)
        if annuity[0] == "immediate":
            premium = self.insurance.minimum_level_premium(annuity = "immediate", interval = annuity[1])
            self.annuity = TermLifeAnnuity(self.insurance.x, self.insurance.term, \
                                           [Premium(self.insurance.x + (i+1)*dt, premium) for i in range(n_payments)], \
                                           self.insurance.interest_rates, \
                                           self.insurance.table)
        elif annuity[0] == "due":
            premium = self.insurance.minimum_level_premium(annuity = "due", interval = annuity[1])
            self.annuity = TermLifeAnnuity(self.insurance.x, self.insurance.term, \
                                           [Premium(self.insurance.x + (i)*dt, premium) for i in range(n_payments)], \
                                           self.insurance.interest_rates, \
                                           self.insurance.table)
        elif type(annuity) != list:
            self.annuity = annuity

    def income_expense(self, t_end):
        income = self.annuity.expected_payments(t_end)
        expense = self.insurance.expected_expense(t_end)
        return income, expense

    def income_expense_df(self, t_end):
        income, expense = self.income_expense(t_end)
        income_df = pandas.DataFrame([(i,j) for i,j in zip(income['t'], income['expected'])], columns = ["t", "income"])
        expense_df = pandas.DataFrame([(i,j) for i,j in zip(expense['t'], expense['expected'])], columns = ["t", "expense"])
        return income_df, expense_df

    def cashflow_df(self, t_end, frequency_weight = 1, collective_interest_rates = None):
        income_df, expense_df = self.income_expense_df(t_end)
        if (collective_interest_rates == None):
            collective_interest_rates = [fm.InterestRate(0, 0, "annual")]
        collective_premiums = [p.collective_fund_premium(False) for p in self.annuity.payments]
        collective_fund_annuity = TermLifeAnnuity(self.insurance.x, self.insurance.term, collective_premiums, \
                                                  collective_interest_rates, self.insurance.table)        
        collective_fund_annuity_df = collective_fund_annuity.expected_payments_df(t_end)
        collective_fund_annuity_df = collective_fund_annuity_df.rename(columns = {'income': 'income_from_collective_fund'})
        df=income_df.merge(expense_df, how='outer', on='t', sort=True)
        df=df.merge(collective_fund_annuity_df, how='outer', on='t', sort=True)
        df=df.fillna(0)
        df.income_from_collective_fund *= frequency_weight
        cashflow = df.income.subtract(df.expense, fill_value=0)
        cashflow_df = pandas.DataFrame([[i,j*frequency_weight] for i,j in zip(df.t, cashflow)], \
                                       columns = ['t', 'profit_exclude_collective_fund'])
                
        cashflow_df = cashflow_df.merge(df, how='outer', on='t', sort=True)
        cashflow_df.fillna(0)
        cashflow_df.income = cashflow_df.income.multiply(frequency_weight)
        cashflow_df.expense = cashflow_df.expense.multiply(frequency_weight)

        exp_contributions = [fm.Contribution(cashflow_df.t[i], cashflow_df.profit_exclude_collective_fund[i]) for i in range(len(cashflow_df))]
        exp_contributions_fund = [fm.Contribution(cashflow_df.t[i], cashflow_df.income_from_collective_fund[i]) for i in range(len(cashflow_df))]

        accumulated = []
        for i in range(1, len(exp_contributions)+1):
            involved = exp_contributions[0:i]
            total = sum([j.accumulate(involved[-1].t, self.interest_rates)  for j in involved])
            accumulated.append(total)

        accumulated_2 = []
        for i in range(1, len(exp_contributions_fund)+1):
            involved = exp_contributions_fund[0:i]
            total = sum([j.accumulate(involved[-1].t, collective_interest_rates)  for j in involved])
            accumulated_2.append(total)
 
        cashflow_df['accumulated_profit_exclude_collective_fund'] = pandas.Series(accumulated, index = cashflow_df.index)
        cashflow_df['accumulated_income_collective_fund'] = pandas.Series(accumulated_2, index = cashflow_df.index)
        cashflow_df['total_accumulated_profit'] = cashflow_df['accumulated_profit_exclude_collective_fund'] + cashflow_df['accumulated_income_collective_fund']

        t_policy = [round(self.t_policy,3)]
        for i in range(len(cashflow_df['t'])-1):
            t_policy.append(round(t_policy[-1] + cashflow_df['t'][i+1]-cashflow_df['t'][i], 3))
        cashflow_df['t_policy'] = pandas.Series(t_policy, index = cashflow_df.index)
        cols = list(cashflow_df.columns)
        cols.insert(1, cols.pop())
        cashflow_df = cashflow_df.reindex(columns=cols)
        return cashflow_df

    def reserves_at(self, t):
        return self.insurance.calculate_apv(point = t) - self.annuity.calculate_apv(point = t)

    def reserves_dynamics(self, dt, t_end):
        N = math.floor((t_end-self.insurance.x)/dt)
        t = [self.insurance.x + dt*i for i in range(N+1)]
        if t[-1]!=t_end:
            t.append(t_end)
        reserves = [self.reserves_at(i) for i in t]
        return t, reserves
    
    def reserves_dynamics_df(self, dt, t_end):
        t, reserves = self.reserves_dynamics(dt, t_end)
        df = pandas.DataFrame(data = [[i,j] for i,j in zip(t, reserves)], \
                              columns = ["t", "Reserves"])
        return df



import random

import pandas


def clean_duplicate_time(list_of_objects):
    new = []; t_taken = []
    for i in range(len(list_of_objects)):
        if list_of_objects[i].t not in t_taken:
            new.append( list_of_objects[i] )
            t_taken.append(list_of_objects[i].t)
    return new


class Contribution(object):

    def __init__(self, t, amount):
        self.t = t
        self.amount = amount

    def filter_interest_rates(self, interest_rates, t_target):
        interest_rates = clean_duplicate_time(interest_rates)
        interest_rates = sorted(interest_rates, key = lambda x: x.t)
        if interest_rates[0].t != 0:
            interest_rates[0].t = 0
        if t_target > self.t:
            while True:
                if len(interest_rates)>1:
                    if (interest_rates[0].t <= self.t < interest_rates[0+1].t):
                        interest_rates = [i for i in interest_rates if i.t < t_target]
                        break
                    else:
                        interest_rates.pop(0)
                else:
                    break
        elif t_target < self.t:
            while True:
                if len(interest_rates)>1:
                    if (interest_rates[0].t <= t_target < interest_rates[0+1].t):
                        interest_rates = [i for i in interest_rates if i.t < self.t]
                        break
                    else:
                        interest_rates.pop(0)
                else:
                    break
        else:
            return []
        return interest_rates

    def accumulate(self, t_end, interest_rates):
        if t_end >= self.t:
            result = self.amount
            interest_rates = self.filter_interest_rates(sorted(interest_rates, key = lambda x: x.t), t_end)
            if len(interest_rates)>0:
                t_start = self.t
                for i in range(len(interest_rates)-1):
                    delta = (interest_rates[i+1].t - t_start)
                    result = interest_rates[i].apply(result, delta)
                    t_start = interest_rates[i+1].t

                delta = t_end-t_start
                result = interest_rates[-1].apply(result, delta)
            return result
        else:
            return 0

    def value_at(self, t, interest_rates):
        if t > self.t:
            return self.accumulate(t, interest_rates)
        elif t == self.t:
            return self.amount
        else:
            result = self.amount
            interest_rates = self.filter_interest_rates(sorted(interest_rates, key = lambda x: x.t), t)
            interest_rates = sorted(interest_rates, key = lambda x: x.t, reverse = True)
            if len(interest_rates)>0:
                t_start = self.t
                for i in range(len(interest_rates)-1):
                    delta = (interest_rates[i].t - t_start)
                    result = interest_rates[i].apply(result, delta)
                    t_start = interest_rates[i].t
                delta = t-t_start
                result = interest_rates[-1].apply(result, delta)
            return result
            

class InterestRate(object):

    periods = {'annual':1, \
               'semi-annual':1/2, \
               'quarter':1/4, \
               'month':1/12}

    @staticmethod
    def get_period_length(period_desc):
        if period_desc in InterestRate.periods:
            return InterestRate.periods[period_desc]
        return period_desc
    
    def __init__(self, t, rate, period_desc, compound = True, discount = False):
        self.t = t
        self.rate = rate
        self.period_desc = period_desc
        self.compound = compound
        self.discount = discount
        self.period_length = self.get_period_length(period_desc)
        
    def apply(self, amount, delta):
        power = delta/self.period_length
        if self.discount:
            return amount*((1-self.rate)**(-power))
        if self.compound:
            return amount*((1+self.rate)**power)
        return amount*(1+(power*self.rate))


class FinancialTL(object):

    def __init__(self, contributions, interest_rates):
        self.contributions = contributions
        self.interest_rates = interest_rates
        self.adjustments()
        
    def adjustments(self):
        self.contributions = sorted(self.contributions, key = lambda x: x.t)
        self.interest_rates = clean_duplicate_time(self.interest_rates)
        self.interest_rates = sorted(self.interest_rates, key = lambda x: x.t)
        if self.interest_rates[0].t != 0:
            self.interest_rates[0].t = 0
        
    def acc_value_at_point(self, t):
        return sum([\
            c.accumulate(t, self.interest_rates)\
            for c in self.contributions])

    def acc_value_dynamics(self, t_end, dt):
        n = round((t_end)/dt)
        t = [(t_end)*i/n for i in range(n+1)]
        acc_values = [self.acc_value_at_point(i) for i in t]
        if t_end not in t:
            t.append(t_end)
            acc_values.append(self.acc_value_at_point(t_end))
        return t, acc_values

    def acc_value_dynamics_df(self, t_end, dt):
        t, acc_values = self.acc_value_dynamics(t_end, dt)
        df = pandas.DataFrame(data = [[i,j] for i,j in zip(t, acc_values)], \
                              columns = ["t", "accum. value (A(t))"])
        return df        

    def value_at(self, t):
        return sum([\
            c.value_at(t, self.interest_rates)\
            for c in self.contributions])


class Bond(object):

    periods = {'annual':1, \
               'semi-annual':1/2, \
               'quarter':1/4, \
               'month':1/12}

    @staticmethod
    def get_period_length(period_desc):
        if period_desc in Bond.periods:
            return Bond.periods[period_desc]
        return period_desc

    def __init__(self, t_start, term, face_value, nominal_coupon_rate, coupon_period, \
                 redemption_value, interest_rates):
        self.t_start = t_start
        self.term = term
        self.face_value = face_value
        self.nominal_coupon_rate = nominal_coupon_rate
        self.coupon_period = coupon_period
        self.coupon_period_length = self.get_period_length(coupon_period)
        self.coupon_rate = self.nominal_coupon_rate*self.coupon_period_length
        self.n_coupons = int(self.term/self.coupon_period_length)
        self.coupon_amount = self.face_value*self.coupon_rate
        self.redemption_value = redemption_value
        self.interest_rates = interest_rates
        self.adjustments()

    def adjustments(self):
        self.interest_rates = clean_duplicate_time(self.interest_rates)
        self.interest_rates = sorted(self.interest_rates, key = lambda x: x.t)
        if self.interest_rates[0].t != 0:
            self.interest_rates[0].t = 0

    @property
    def coupon_payments(self):
        payments = [Contribution(self.t_start+self.coupon_period_length*i, \
                                  self.coupon_amount) for i in range(1, self.n_coupons+1)]
        return payments

    @property
    def price(self):
        redemption = Contribution(self.t_start + self.term, self.redemption_value)
        timeline = FinancialTL(self.coupon_payments + [redemption], \
                               self.interest_rates)
        return timeline.value_at(self.t_start)  


class Installments(object):

    def __init__(self, interval, principle_amount, payments, interest_rates, do_adjustments = True):
        self.interval = interval
        self.t_start = interval[0]
        self.t_end = interval[1]
        self.principle = Contribution(self.t_start, principle_amount)
        self.payments = sorted(payments, key = lambda x: x.t)
        self.interest_rates = sorted(interest_rates, key = lambda x: x.t)
        if do_adjustments:
            self.adjustments()
        
    def adjustments(self):
        payments = []; t_taken = []
        for i in range(len(self.payments)):
            if self.payments[i].t not in t_taken:
                payments.append( self.payments[i] )
                t_taken.append(self.payments[i].t)
        self.payments = payments
        
        int_rates = []; t_taken = []
        for i in range(len(self.interest_rates)):
            if self.interest_rates[i].t not in t_taken:
                int_rates.append( self.interest_rates[i] )
                t_taken.append(self.interest_rates[i].t)
        self.interest_rates = int_rates
        if 0 not in t_taken:
            self.interest_rates[0].t = 0

    def minimum_required_single_payment(self, t_payment):
        if t_payment >= self.t_start:
            unit_contribution = Contribution(t_payment, 1)
            unit_present_value = unit_contribution.value_at(self.t_start, self.interest_rates)
            return self.principle.amount/unit_present_value
        else:
            return None

    def outstanding_balance_at_point(self, t):
         accum_payments = sum([p.accumulate(t, self.interest_rates)\
                               for p in self.payments])
         loan_amount = self.principle.accumulate(t, self.interest_rates)
         return loan_amount - accum_payments

    def outstanding_balance_at_end(self):
        return self.outstanding_balance_at_point(self.t_end)

    def outstanding_balance_dynamics(self, dt):
        n = round((self.t_end-self.t_start)/dt)
        t = [self.t_start + (self.t_end-self.t_start)*i/n for i in range(n+1)]
        balances = [self.outstanding_balance_at_point(i) for i in t]
        if self.t_end not in t:
            t.append(self.t_end)
            balances.append(self.outstanding_balance_at_end())
        return t, balances

    def outstanding_balance_df(self, dt):
        t, balances = self.outstanding_balance_dynamics(dt)
        df = pandas.DataFrame(data = [[i,j] for i,j in zip(t, balances)], \
                              columns = ["t", "Outstanding Balance"])
        return df



import setuptools


setuptools.setup(\
    name='actuarydesk',\
    version='0.0.1',\
    author='Arief Anbiya',\
    author_email='anbarief@live.com',\
    description='Tools for actuary to model and analyze simple products or for actuarial students to practice.',\
    url='https://github.com/pypa/actuarydesk',\
    packages=setuptools.find_packages(),\
    classifiers=["Programming Language :: Python :: 3",\
                 "License :: OSI Approved :: MIT License",\
                 "Operating System :: OS Independent"],\
    python_requires='>=3.6')
